from flask import Flask, jsonify
from threading import Thread
from functools import reduce
from keras.datasets import boston_housing
from keras import models, layers, optimizers
from keras import backend as K
from random import uniform
from numpy import linalg
from scipy.misc import derivative
import flask
import numpy as np
import os, signal, sys
import urllib.request
import urllib.parse
import argparse
import random
import time
import json
import keras
import os.path
import math


class Logger(object):
    def __init__(self, prefix='Global'):
        self.prefix = prefix

    def write(self, message):
        formatted_message = '[Sensei] (' + self.prefix + '): ' + message
        print(formatted_message)
        pass

    def important(self, message):
        formatted_message = '[Sensei] <IMPORTANT> (' + self.prefix + '): ' + message
        print(formatted_message)


class Arguments(object):
    def __init__(self):
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument('--type', help='Job name, either ps or worker', required=True)
        self.parser.add_argument('--task_index', help='Task index, 0 when ps, i when worker', required=True, type=int)
        self.parser.add_argument('--worker', action='append', help='Worker address', required=True)
        self.parser.add_argument('--server', help='Server address', required=True)

    def get_arguments(self):
        arguments = self.parser.parse_args()
        return arguments.__dict__


class URLThread(Thread):
    def __init__(self, url, body=None, retry=False):
        super(URLThread, self).__init__()
        self.logger = Logger(f'URLThread {url}')
        self.address = url
        self.retry = retry
        self.response = None
        self.body = body

    def run(self):
        self.logger.write('Sending GET request')
        self.response = self.get()
        self.logger.write('Done.')

    def get(self):
        while True:
            try:
                byte_body = None if self.body is None else urllib.parse.urlencode(self.body).encode("utf-8")
                request = urllib.request.Request(self.address, data=byte_body)
                self.logger.write(f'Sending request with {self.body}')
                with urllib.request.urlopen(request) as response:
                    message = response.read()
                    self.logger.write(f'Received response: {message}')
                    return message
            except Exception as error:
                if not self.retry:
                    self.logger.write(f'Error occured, decided not to retry: {error}')
                    return 'Error'
                self.logger.write(f'Error occured, retrying: {error}')
                time.sleep(2)


class ConcurrentHttpClient(object):
    def __init__(self):
        self.logger = Logger('Concurrent HTTP client')

    def get(self, urls, bodies=None, retry=False):
        def alive_count(threads):
            alive = map(lambda request_thread: 1 if request_thread.isAlive() else 0, threads)
            return reduce(lambda left, right: left + right, alive)

        count = len(urls)
        self.logger.write(f'Creating {count} threads.')
        requests = []
        for index in range(len(urls)):
            body = None if bodies is None else bodies[index]
            thread = URLThread(urls[index], body=body, retry=retry)
            requests.append(thread)

        self.logger.write(f'Starting {count} request threads.')
        for request in requests:
            request.start()

        self.logger.write(f'Waiting for {count} threads to exit...')
        while alive_count(requests) > 0:
            time.sleep(1)

        self.logger.write(f'Successfully managed for {count} requests to complete.')
        return [request.response for request in requests]


class Evolve(object):
    def __init__(self):
        self.logger = Logger('Evolve')
        self.matrix = {
            'x0': [index for index in range(10, 100, 4)],
            'x1': [index for index in range(10, 100, 4)],
            'x2': ['relu', 'tanh', 'sigmoid', 'softmax'],
            'x3': ['relu', 'tanh', 'sigmoid', 'softmax'],
            'x4': [1e-4, 1e-3, 1e-2, 1e-1],
            'x5': [optimizers.RMSprop, optimizers.SGD, optimizers.Adam]
        }

    def score(self, point):
        decoded = self.decode_parameters(point)
        score = self.k_blocks_score(decoded)
        return score

    def decode_parameters(self, point):
        self.logger.write(f'Decoding parameters from {point}')
        integers = np.rint(point).astype(int)
        decoded = {}
        for index, integer in enumerate(integers):
            key = f'x{index}'
            matrix_element = self.matrix[key]
            parameter = matrix_element[integer]
            decoded[key] = parameter
        return decoded

    def k_blocks_score(self, decoded, k=2):
        self.logger.write('Loading and splitting Boston Housing dataset...')
        x_train, y_train, x_test, y_test = self.load_dataset()
        num_samples = len(x_train) // k
        all_scores = []
        self.logger.write('Performing k-blocks check...')
        for index in range(k):
            self.logger.write(f'Measuring model for {index}-th block...')
            lower = index * num_samples
            higher = (index + 1) * num_samples
            mse, mae = self.build_model(
                x_train=np.concatenate([x_train[:lower], x_train[higher:]], axis=0),
                y_train=np.concatenate([y_train[:lower], y_train[higher:]], axis=0),
                x_test=x_train[lower:higher],
                y_test=y_train[lower:higher],
                neurons_first=decoded['x0'],
                neurons_second=decoded['x1'],
                activation_first=decoded['x2'],
                activation_second=decoded['x3'],
                learning_rate=decoded['x4'],
                optimizer=decoded['x5'])
            self.logger.write(f'MSE: {mse}; MAE: {mae}')
            all_scores.append(mae)
        mean_score = np.mean(all_scores)
        self.logger.write(f'Successfully obtained score: {mean_score}')
        if np.isnan(mean_score):
            return 99
        return mean_score

    @staticmethod
    def load_dataset():
        (x_train, y_train), (x_test, y_test) = boston_housing.load_data()
        mean = x_train.mean(axis=0)
        std = x_train.std(axis=0)
        x_train -= mean
        x_train /= std
        x_test -= mean
        x_test /= std
        return x_train, y_train, x_test, y_test

    @staticmethod
    def build_model(
            x_train,
            y_train,
            x_test,
            y_test,
            neurons_first,
            neurons_second,
            activation_first,
            activation_second,
            learning_rate,
            optimizer):
        K.clear_session()
        model = models.Sequential()
        model.add(layers.Dense(neurons_first, activation=activation_first, input_shape=(x_train.shape[1],)))
        model.add(layers.Dense(neurons_second, activation=activation_second, input_dim=neurons_first))
        model.add(layers.Dense(1, input_dim=neurons_second))
        model.compile(optimizer=optimizer(lr=learning_rate), loss='mse', metrics=['mae'])
        model.fit(x_train, y_train, epochs=35, batch_size=1, verbose=0)
        mse, mae = model.evaluate(x_test, y_test, verbose=0)
        return mse, mae


class GradientAscentWorker(object):
    def __init__(self, func, higher, lower, epsilon=1):
        self.logger = Logger('Gradient Ascent Worker')
        self.epsilon = epsilon
        self.higher = higher
        self.lower = lower
        self.func = func

    def bounded(self, point):
        return np.maximum(
            np.minimum(np.array(point), self.higher),
            self.lower
        ).tolist()

    def partial_derivative(self, index, point):
        self.logger.write(f'Computing partial derivative for {len(point)} dimensions, var: {index}')
        arguments = point[:]
        def substitute(x):
            arguments[index] = x
            bounded_arguments = self.bounded(arguments)
            return self.func(*bounded_arguments)
        return derivative(substitute, point[index], dx=self.epsilon, order=3)


class Worker(object):
    def __init__(self, worker_index, workers):
        address = workers[worker_index]
        self.evolve = Evolve()
        self.host = address.split(':')[0]
        self.port = address.split(':')[1]
        self.worker_index = worker_index
        self.worker_count = len(workers)
        self.logger = Logger(f'Worker {address}')
        self.allowed_methods = ['GET', 'POST']
        self.app = Flask(__name__)
        self.function = lambda args: -sum([x ** 2 for x in args])

    def ping(self):
        self.logger.write('Received /ping request.')
        return 'Pong!'

    def compute(self):
        self.logger.write('Received /compute request.')
        points = json.loads(flask.request.form["points"])
        scores = [-self.evolve.score(point) for point in points]
        self.logger.write(f'Obtained scores: {scores}')
        return str(scores)
    
    def derivative(self):
        self.logger.write('Received /derivative request.')
        point = json.loads(flask.request.form["point"])
        indexes = json.loads(flask.request.form["indexes"])
        self.logger.write(f'Computing indexes {indexes}, for point {point}')
        
        worker = GradientAscentWorker(
            func=lambda *args: -self.evolve.score(args),
            lower=np.array([0, 0, 0, 0, 0, 0]),
            higher=np.array([20, 20, 3, 3, 3, 2]))
        
        derivatives = [worker.partial_derivative(indexes[i], point) for i in range(len(indexes))]
        self.logger.write(f'Obtained derivative: {derivatives}')
        return str(derivatives)

    def shutdown(self):
        self.logger.write(f'Received /shutdown request.')
        os.kill(os.getpid(), signal.SIGTERM)
        return 'Shutting down the app.'

    def run(self):
        self.logger.write(f'Starting Flask app...')
        self.app.route('/ping', methods=self.allowed_methods)(self.ping)
        self.app.route('/compute', methods=self.allowed_methods)(self.compute)
        self.app.route('/shutdown', methods=self.allowed_methods)(self.shutdown)
        self.app.route('/derivative', methods=self.allowed_methods)(self.derivative)
        self.app.run(host='0.0.0.0', port=self.port, debug=True)


class FishSchoolSearch(object):
    def __init__(self, compute, lower, higher, iter_max=10, step_ind=10, step_vol=5, count=8):
        self.logger = Logger('Fish School Search')
        self.logger.write(f'Running {iter_max} steps, spawning {count} agents, using {len(lower)} dimensions.')
        self.step_ind_initial = step_ind
        self.step_vol_initial = step_vol
        self.step_ind = step_ind
        self.step_vol = step_vol
        self.iter_max = iter_max
        self.compute = compute
        self.higher = higher
        self.lower = lower
        self.count = count
        self.steps = []
        self.total_weight = 0
        self.points = [self.random_bounded_tensor() for i in range(count)]
        self.point_deltas = [np.zeros(x.shape) for x in self.points]
        self.score_deltas = [np.zeros(1) for x in self.points]
        self.weights = [np.zeros(1) for x in self.points]
        self.fitnesses = self.compute(self.points)
        self.best_fitness = self.fitnesses[0]
        self.best_point = self.points[0]
        self.update_best()

    def random_bounded_tensor(self):
        empty = np.empty(self.lower.shape)
        for index in range(self.lower.shape[0]):
            empty[index] = uniform(self.lower[index], self.higher[index])
        return empty

    def bounded(self, point):
        return np.maximum(np.minimum(point, self.higher), self.lower)

    def individual_move(self):
        updated_points = []
        for index in range(self.count):
            displacement = self.step_ind * np.random.uniform(-1, 1, self.lower.shape)
            current_point = self.points[index]
            updated_point = self.bounded(current_point + displacement)
            updated_points.append(updated_point)
        updated_fitnesses = self.compute(updated_points)
        for index in range(self.count):
            current_point = self.points[index]
            current_score = self.fitnesses[index]
            updated_point = updated_points[index]
            updated_score = updated_fitnesses[index]
            if updated_score >= current_score:
                self.points[index] = updated_point
                self.point_deltas[index] = updated_point - current_point
                self.score_deltas[index] = updated_score - current_score
                self.fitnesses[index] = updated_score
            else:
                self.point_deltas[index] = np.zeros(current_point.shape)
                self.score_deltas[index] = np.zeros(current_score.shape)

    def feed(self, max_weight=300):
        maximum_delta = np.max(np.stack(self.score_deltas))
        for index in range(self.count):
            weight = self.weights[index]
            score_delta = self.score_deltas[index]
            delta = 0 if maximum_delta == 0 else score_delta / maximum_delta
            updated_weight = weight + delta
            if updated_weight > max_weight:
                updated_weight = max_weight
            self.weights[index] = updated_weight

    def instinctive_move(self):
        scalar_sum = 0
        vector_sum = np.zeros(self.lower.shape)
        for index in range(self.count):
            score_delta = self.score_deltas[index]
            vector_sum += self.point_deltas[index] * score_delta
            scalar_sum += score_delta
        i = np.zeros(self.lower.shape)
        if scalar_sum != 0:
            i = vector_sum / scalar_sum
        for index in range(self.count):
            current_point = self.points[index]
            self.points[index] = self.bounded(current_point + i)

    def barycenter_move(self):
        vector_sum = np.zeros(self.lower.shape)
        scalar_sum = 0
        for index in range(self.count):
            weight = self.weights[index]
            vector_sum += self.points[index] * weight
            scalar_sum += weight
        b = 0 if scalar_sum == 0 else vector_sum / scalar_sum
        total_weight = np.sum(np.stack(self.weights))
        for index in range(self.count):
            point = self.points[index]
            step = self.step_vol * np.random.uniform(0, 1, self.lower.shape)
            sub = point - b
            norm = np.linalg.norm(sub)
            diff = step * sub / (1 if norm == 0 else norm)
            if total_weight > self.total_weight:
                self.points[index] = self.bounded(point - diff)
            else:
                self.points[index] = self.bounded(point + diff)
        self.total_weight = total_weight

    def update_progress(self, iteration):
        progress = (self.iter_max - iteration) / self.iter_max
        self.step_ind = self.step_ind_initial * progress
        self.step_vol = self.step_vol_initial * progress

    def update_best(self):
        for index in range(self.count):
            point = self.points[index]
            current_fitness = self.fitnesses[index]
            if current_fitness > self.best_fitness:
                self.best_fitness = current_fitness
                self.best_point = point
        self.steps.append(self.best_fitness)

    def step(self, iteration):
        self.individual_move()
        self.feed()
        self.instinctive_move()
        self.barycenter_move()
        self.update_best()
        self.update_progress(iteration)

    def run(self):
        self.logger.write('Starting Fish School Search...')
        for index in range(self.iter_max):
            self.logger.write(f'Running step {index}...')
            self.step(index + 1)
        self.logger.write('Completing Fish School Search...')
        return self.best_point, self.steps


class GradientAscentServer(object):
    def __init__(self, func, gradient, start, lower, higher, gamma=1, limit=3):
        self.logger = Logger('Gradient Ascent Server')
        self.gradient = gradient
        self.start = start
        self.gamma = gamma
        self.limit = limit
        self.higher = higher
        self.lower = lower
        self.func = func

    def bounded(self, point):
        return np.maximum(
            np.minimum(np.array(point), self.higher), 
            self.lower
        ).tolist()

    def run(self):
        iteration = 0
        gamma = self.gamma
        current = self.start
        self.logger.write(f'Running {self.limit} iterations')
        steps = [self.func(*current)]
        while iteration < self.limit:
            iteration += 1
            components = []
            for component in self.gradient(current):
                if component > 0:
                    value = math.ceil(component)
                    components.append(value)
                else:
                    value = math.floor(component)
                    components.append(value)
            self.logger.write(f'Gradient value on {iteration}-th iteration is: {components}')
            grad = np.array(components) * gamma
            current = self.bounded(np.add(current, grad).tolist())
            steps.append(self.func(*current))
        return current, steps


class Server(object):
    def __init__(self, workers):
        self.distributed_http = ConcurrentHttpClient()
        self.logger = Logger('Server')
        self.evolve = Evolve()
        self.workers = workers
        self.ping_urls = self.urls('/ping')
        self.compute_urls = self.urls('/compute')
        self.gradient_urls = self.urls('/derivative')
        self.shutdown_urls = self.urls('/shutdown')
        self.points_multiplier = 1

    def run(self):
        self.logger.write('Running distributed workflow...')
        self.distributed_http.get(self.ping_urls, retry=True)
        self.logger.write(f'Successfully managed to wait for {len(self.workers)} workers.')

        self.logger.important('Running the Fish School Search distributed optimizer...')
        start = time.time()
        best_point, all_points = \
            FishSchoolSearch(
                compute=lambda points: self.fitness(points),
                lower=np.array([0, 0, 0, 0, 0, 0]),
                higher=np.array([20, 20, 3, 3, 3, 2]),
                count=len(self.workers) * self.points_multiplier,
                step_ind=4, step_vol=2,
                iter_max=8) \
                .run()

        self.logger.write(f'Successfully managed to complete FSS execution, seconds passed: {time.time() - start}')
        self.logger.important(f'Best solution is: {best_point}')
        self.logger.important(f'All solutions are: {[point.item() for point in all_points]}')
        
        self.logger.important('Running the Gradient Ascent distributed optimizer...')
        start = time.time()
        best_point, all_points = \
            GradientAscentServer(
                func=lambda *args: -self.evolve.score(args),
                gradient=lambda point: self.gradient(point),
                start=best_point.tolist(),
                lower=np.array([0, 0, 0, 0, 0, 0]),
                higher=np.array([20, 20, 3, 3, 3, 2]),
                gamma=1,
                limit=2) \
                .run()

        self.logger.write(f'Successfully managed to complete GD execution, seconds passed: {time.time() - start}')
        self.logger.important(f'Best solution is: {best_point}')
        self.logger.important(f'All solutions are: {all_points}')
        self.logger.write('Shutting down all the workers...')
        self.distributed_http.get(self.shutdown_urls, retry=False)

    def fitness(self, points):
        self.logger.write(f'Computing fitness of {len(points)}')
        request_bodies = []
        for index in range(0, len(points), self.points_multiplier):
            request_points = points[index:index + self.points_multiplier]
            request_points = list(map(lambda n: n.tolist(), request_points))
            request_bodies.append({'points': request_points})

        responses = self.distributed_http.get(self.compute_urls, bodies=request_bodies, retry=False)
        self.logger.write(f'Received {len(responses)} responses: {responses}')
        
        solutions = []
        for byte_list in responses:
            decoded_list = json.loads(byte_list)
            for solution in decoded_list:
                np_value = np.array(solution)
                solutions.append(np_value)
        
        self.logger.write(f'Mapped responses into {len(solutions)} solutions: {solutions}')
        return solutions
    
    def gradient(self, point):
        self.logger.write(f'Computing gradient of point {point}')
        worker_count = len(self.workers)
        dimensionality = len(point)
        global_responses = []
        for batch_index in range(0, dimensionality, worker_count):
            request_bodies = []
            for index in range(0, worker_count):
                global_index = index + batch_index
                if global_index < dimensionality:
                    request = {'point': point, 'indexes': [index]}
                    request_bodies.append(request)

            urls = self.gradient_urls[:len(request_bodies)]
            self.logger.write(f'Sending {len(request_bodies)} requests to {len(urls)} workers...')
            responses = self.distributed_http.get(urls, bodies=request_bodies, retry=False)

            self.logger.write(f'Received {len(responses)} responses: {responses}')
            for response in responses:
                global_responses.append(response)
        
        self.logger.write(f'Post-processing {len(global_responses)} responses...')
        gradients = []
        for byte_list in global_responses:
            decoded_list = json.loads(byte_list)
            for gradient in decoded_list:
                gradients.append(gradient)

        self.logger.write(f'Mapped responses into {len(gradients)}-dim gradient vector: {gradients}')
        return gradients

    def urls(self, path):
        lazy = map(lambda address: f'http://{address}{path}', self.workers)
        return list(lazy)


if __name__ == '__main__':
    parameters = Arguments().get_arguments()
    if parameters['type'] != 'ps':
        Worker(parameters['task_index'], parameters['worker']).run()
    else:
        Server(parameters['worker']).run()